<?php

declare(strict_types=1);

namespace app\controllers;

use yii;
use yii\filters\AccessControl;
use yii\web\Controller;
use yii\web\Response;

use app\models\Order;
use app\models\AccountEdgeOrder;
use Exception;

class OrderController extends Controller
{
    public function behaviors()
    {
        return [
            'access' => [
                'class' => AccessControl::class,
                'rules' => [
                    [
                        'allow' => true,
                        'roles' => ['@'],
                        'actions' => ['index', 'write', 'delete', 'amount-update', 'pay']
                    ],
                    [
                        'allow' => false,
                        'roles' => ['?'],
                        'denyCallback' => [$this, 'accessDenied']
                    ]
                ]
            ]
        ];
    }

    public function accessDenied()
    {
        // Инициализация
        $cookies = yii::$app->response->cookies;

        // Запись cookie с редиректом, который выполнится после авторизации
        $cookies->add(new Cookie([
            'name' => 'redirect',
            'value' => yii::$app->request->pathInfo
        ]));

        if (Yii::$app->request->isPost) {
            // POST-запрос

            // Настройка
            Yii::$app->response->format = Response::FORMAT_JSON;

            // Генерация ответа
            Yii::$app->response->content = json_encode([
                'main' => $this->renderPartial('/orders/index'),
                'redirect' => '/authentication',
                '_csrf' => Yii::$app->request->getCsrfToken()
            ]);
        } else if (Yii::$app->request->isGet) {
            // GET-запрос

            $this->redirect('/authentication');
        }
    }

    public function actionIndex()
    {
        // Инициализация
        $orders = Order::search(type: 'all', limit: 10, page: 1, select: '{account_edge_order, order}');

        if (yii::$app->request->isPost) {
            // POST-запрос

            yii::$app->response->format = Response::FORMAT_JSON;

            return [
                'main' => $this->renderPartial('/orders/index', compact('orders')),
                'title' => 'Заказы',
                'redirect' => '/orders',
                '_csrf' => yii::$app->request->getCsrfToken()
            ];
        }

        return $this->render('/orders/index', compact('orders'));
    }

    /**
     * Запись
     */
    public function actionWrite(): string|array|null
    {
        if (yii::$app->request->isPost) {
            // POST-запрос

            // Инициализация входных данных
            $account = yii::$app->user;
            $supplies = yii::$app->request->post('supplies');

            yii::$app->response->format = Response::FORMAT_JSON;

            // Инициализация возврата по умолчанию
            $return = [
                '_csrf' => yii::$app->request->getCsrfToken()
            ];

            if (is_null($supplies)) {
                // 501 Not Implemented
                yii::$app->response->statusCode = 501;
            }

            if (!is_null($supplies)) {
                // Переданы поставки для записи

                if (!$model = Order::search($account)) {
                    // Корзина не найдена (текущий заказ)

                    // Инициализация
                    $model = new Order();
                    $model->save() or throw new Exception('Не удалось инициализировать заказ');

                    // Запись ребра: АККАУНТ -> ЗАКАЗ
                    AccountEdgeOrder::write($account->id, $model->readId(), 'current') or $model->addError('errors', 'Не удалось инициализировать ребро: АККАУНТ -> ЗАКАЗ');
                }

                // Проверка входных данных
                if (!is_array($supplies)) {
                    // Неверные входные данные

                    // Запись ошибки
                    $model->addError('errors', 'Переменная должна быть массивом');
                }

                // Если запись не удалась, то вернуть код: 500 Internal Server Error
                $model->writeSupply($supplies) or yii::$app->response->statusCode = 500;
            }

            return $return;
        }

        yii::$app->response->statusCode = 500;

        return null;
    }

    /**
     * Удаление
     *
     * @todo Разделить логику (для удобства чтения)
     */
    public function actionDelete(): string|array|null
    {
        // Инициализация
        $targets = yii::$app->request->post('targets') ?? yii::$app->request->get('targets');
        $page = yii::$app->request->get('page') ?? yii::$app->request->post('page') ?? 1;
        $account = yii::$app->user;
        $order = Order::search();

        if ($targets) {
            // Удаление выбранных целей (удаление из заказа)

            foreach (isset($targets[0]) && is_array($targets[0]) ? $targets : [$targets] as $target) {
                // Унификация входных параметров

                foreach ($target as $catn => $amount) {
                    // Перебор целей

                    // Удаление
                    $order->deleteSupply([$catn => (int) $amount]);
                }
            }
        } else {
            // Целью подразумевается сам заказ (удаление заказа)

            // Инициализация
            $order->stts = 'reserved';

            if ($order->update()) {
                // Запись в журнал
                $order->journal('reserved');

                // Поиск
                $edge = AccountEdgeOrder::searchByVertex($account->id, $order->readId(), 'current');

                if (count($edge) > 1) {
                    // Найден более чем 1 заказ

                    return null;
                }

                // Инициализация
                $edge = $edge[0];
                $edge->type = 'reserved';

                // Запись
                $edge->update();

                // Реинициализация
                $order = Order::search();

                if (empty($order)) {
                    // Корзина не инициализирована

                    // Инициализация
                    $order = new Order();
                    $order->save() or throw new Exception('Не удалось инициализировать заказ');

                    // Подключение
                    $order->connect($account);
                }
            }
        }

        // Инициализация содержимого корзины
        $supplies = $order->content(10, $page);

        if (yii::$app->request->isPost) {
            // POST-запрос

            yii::$app->response->format = Response::FORMAT_JSON;

            return [
                'main' => $this->renderPartial('/cart/index', compact('order', 'supplies')),
                'title' => 'Корзина',
                'redirect' => '/cart',
                '_csrf' => yii::$app->request->getCsrfToken()
            ];
        }

        return $this->render('/cart/index', compact('order', 'supplies'));
    }

    /**
     * Обновление количества товара
     */
    public function actionAmountUpdate(): string|array|null
    {
        // Инициализация
        $targets = yii::$app->request->post('targets') ?? yii::$app->request->get('targets');
        $page = yii::$app->request->get('page') ?? yii::$app->request->post('page') ?? 1;
        $account = yii::$app->user;
        $order = Order::search();
        $supplies = $order->content(10, $page);

        foreach (isset($targets[0]) && is_array($targets[0]) ? $targets : [$targets] as $target) {
            // Унификация входных параметров

            foreach ($target as $catn => $amount) {
                // Перебор целей (переданных объектов в корзине)

                foreach ($supplies as $supply) {
                    // Перебор объектов в корзине

                    if ($supply->catn === $catn) {
                        // Цель найдена

                        if ($supply->amnt > $amount) {
                            // Запрошено уменьшение количества

                            // Удаление
                            $order->deleteSupply([$catn => $supply->amnt - $amount]);
                        } else if ($supply->amnt < $amount) {
                            // Запрошено увеличение количества

                            // Запись
                            $order->writeSupply([$supply->catn => $amount - $supply->amnt]);
                        }
                    }
                }
            }
        }

        // Ренициализация
        $supplies = $order->content(10, $page);

        if (yii::$app->request->isPost) {
            // POST-запрос

            yii::$app->response->format = Response::FORMAT_JSON;

            return [
                'main' => $this->renderPartial('/cart/index', compact('order', 'supplies')),
                'title' => 'Корзина',
                'redirect' => '/cart',
                '_csrf' => yii::$app->request->getCsrfToken()
            ];
        }

        return $this->render('/cart/index', compact('order', 'supplies'));
    }

    /**
     * Оплата
     */
    public function actionPay(): string|array|null
    {
        // Инициализация
        $model = Order::search();

        // Поиск ребра
        $edge = AccountEdgeOrder::searchByVertex(yii::$app->user->id, $model->readId(), 'current');

        if (count($edge) > 1) {
            // Найден более чем 1 заказ

            return null;
        }

        // Инициализация
        $edge = $edge[0];

        // Запись
        $edge->type = 'accepted';

        // Отправка изменений
        $edge->update();

        // Инициализация
        $orders = Order::search(type: 'all', limit: 10, page: 1, select: '{account_edge_order, order}');

        if (yii::$app->request->isPost) {
            // POST-запрос

            yii::$app->response->format = Response::FORMAT_JSON;

            return [
                'main' => $this->renderPartial('/orders/index', compact('orders')),
                'title' => 'Заказы',
                'redirect' => '/orders',
                '_csrf' => yii::$app->request->getCsrfToken()
            ];
        }

        return $this->render('/orders/index', compact('orders'));
    }
}
